"
 COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
"{ Package: 'stx:libbasic' }"

CompiledCode variableSubclass:#Method
	instanceVariableNames:'source sourcePosition category package mclass'
	classVariableNames:'PrivateMethodSignal LastFileReference LastSourceFileName
		LastWhoClass LastFileLock LastMethodSources LastMethodSourcesLock
		CompilationLock'
	poolDictionaries:''
	category:'Kernel-Methods'
!

Object subclass:#MethodWhoInfo
	instanceVariableNames:'myClass mySelector'
	classVariableNames:''
	poolDictionaries:''
	privateIn:Method
!

!Method class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    this class defines protocol for executable methods;
    both compiled and interpreted methods are represented by this class.
    Compiled methods have a non-nil code field, while interpreted methods have
    a nil code field and non-nil byteCode field.
    If there are both non-nil code and bytecode fields, the VM will execute
    the machine-code of a method. If both are nil when executed, a noByteCode
    message is sent by the VM to the method where a signal is raised.

    The methods sourcecode is represented by source and sourcePosition:
    - if sourcePosition is a Number, the source-field is the fileName and
      sourcePosition is the character offset of the source-chunk in this source file.
    - If sourcePosition is nil, the source is the string in the source field.
    (an old version used ExternalString instances here, but that lead to
     10000 additional little objects ...)

    The flags field defines things like the number of method-locals,
    method arguments and stack requirements (for interpreted methods).
    Do not depend on any value in the flags field - it may change without
    notice.

    Notice, that in ST/X, method can be subclassed; executable code is
    identified not by being an instance of Block or Method, but instead by
    having the executable flag bit set in the class. The VM can execute anything
    which is identified as executable (assuming that the first instance variable
    is the machine-code address) - this allows for easy future extension.

    [Instance variables:]

        source          <String>        the source itself (if sourcePosition isNil)
                                        or the fileName where the source is found

        sourcePosition  <Integer>       the position of the methods chunk in the file

        category        <Symbol>        the methods category
        package         <Symbol>        the package, in which the methods was defined
        mclass          <Class>         the class in which I am defined
        indexed slots                   literals

    [Class variables:]

        PrivateMethodSignal             raised on privacy violation (see docu)

        LastFileReference               weak reference to the last sourceFile
        LastSourceFileName              to speedup source access via NFS

    WARNING: layout known by compiler and runtime system - dont change

    [author:]
        Claus Gittinger
"
!

dynamicMethods
"
    On systems which support dynamic loading of stc-compiled machine code (SYS5.4, Linux),
    methods may now be compiled to machine code from within the browser,
    and the resulting machine code object be loaded in.
    The ObjectFileLoader keeps (weak) handles to the resulting methods and
    invalidates the corresponding method objects, if the underlying methods
    object code is unloaded.
    Invalid methods will trap into the debugger when executed;
    also, the browser marks them as '(* not executable *)' in its method list.

    Notice: this has nothing to do with JIT compilation, which is always possible.
    JIT-compilation is done from bytecodes to a machineCode cache.
    In contrast, dynamic loading of stc-compiled code goes via intermediate C-code
    which is compiled by the machines native C-compiler.
    As opposed to JITted code, this allows for embedded primitive C-code.
"
!

privacy 
"
    ST/X includes an EXPERIMENTAL implementation of method privacy.
    Individual methods may be set to private or protected via the
    privacy:#private and privacy:#protected messages. Also, categories may be
    filedIn as a whole as private using #privateMethodsFor: or as
    protected using #protectedMethodsFor: instead of the well known #methodsFor:.

    The additional #publicMethodsFor: is for documentation purposes, and
    is equivalent to #methodsFor: (also to support fileIn of ENVY methods).

    Protected methods may be executed only when called via a self-send
    from the superclass-methods and self or super-sends from methods in the
    class itself or subclasses.
    Private methods may not be called from subclasses-methods, 
    i.e. they may only be called via self sends from within the current class.
    (i.e. protected methods are less private than private ones)

    When such a situation arises, the VM (runtime system) will raise the
    PrivateMethodSignal exception (if nonNil), which usually brings you into the
    debugger.

    If PrivateMethodSignal is nil, the VM will not check for this, and
    execution is as usual. (you may want to nil-it for production code,
    and leave it non nil during development).

    NOTICE: there is no (not yet ?) standard defined for method privacy,
    however, the definition protocol was designed to be somewhat ENVY compatible 
    (from what can be deduced by reading PD code).

    Also, the usability of privacy is still to be tested.
    This interface, the implementation and the rules for when a privacy violation
    may change (in case of some ANSI standard being defined).
    Be warned and send me suggestions & critics (constructive ;-)

    Late note (Feb 2000):
        the privacy feature has new been in ST/X for some years and was NOT heavily
        used - neither at eXept, nor by customers.
        In Smalltalk, it seems to be a very questionable feature, actually limiting
        code reusability.
        The privacy features are left in the system to demonstrate that it can be
        done in Smalltalk (for religious C++ fans ... to avoid useless discussions)
        (the check is not expensive, w.r.t. the VM runtime behavior).
"
! !

!Method class methodsFor:'initialization'!

initialize
    "create signals"

    PrivateMethodSignal isNil ifTrue:[
        "EXPERIMENTAL"
        PrivateMethodSignal := ExecutionError newSignalMayProceed:true.
        PrivateMethodSignal nameClass:self message:#privateMethodSignal.
        PrivateMethodSignal notifierString:'attempt to execute private/protected method'.
    ].

    LastFileLock isNil ifTrue:[
        LastFileLock := RecursionLock new name:'Method-LastFile'.
        LastMethodSourcesLock := RecursionLock new name:'Method-LastMethodSources'.
        LastFileReference := WeakArray new:1.
        LastFileReference at:1 put:0.
    ].

    CompilationLock := RecursionLock new name:'MethodCompilation'.

    "Modified: 22.4.1996 / 16:34:38 / cg"
    "Modified: 3.1.1997 / 16:58:16 / stefan"
!

lastMethodSourcesLock
    LastMethodSourcesLock isNil ifTrue:[
        self initialize
    ].
    ^ LastMethodSourcesLock
! !

!Method class methodsFor:'Signal constants'!

privateMethodSignal
    "return the signal raised when a private/protected method is called
     by some other object (i.e. not a self- or super send)"

    ^ PrivateMethodSignal
! !


!Method class methodsFor:'queries'!

isBuiltInClass
    "return true if this class is known by the run-time-system.
     Here, true is returned for myself, false for subclasses."

    ^ self == Method

    "Modified: 23.4.1996 / 15:59:50 / cg"
!

methodDefinitionTemplateForSelector:aSelector
    "given a selector, return a prototype definition string"

    |nA argNames|

    (nA := aSelector numArgs) == 1 ifTrue:[
        argNames := #('arg')
    ] ifFalse:[
        argNames := (1 to:nA) collect:[:i | 'arg' , i printString].
    ].
    ^ self 
        methodDefinitionTemplateForSelector:aSelector 
        andArgumentNames:argNames.

    "
     Method methodDefinitionTemplateForSelector:#foo           
     Method methodDefinitionTemplateForSelector:#+             
     Method methodDefinitionTemplateForSelector:#foo:bar:baz:  
    "
!

methodDefinitionTemplateForSelector:aSelector andArgumentNames:argNames
    "given a selector, return a prototype definition string"

    aSelector numArgs > 0 ifTrue:[
        aSelector isKeyword ifTrue:[
            ^ String streamContents:[:stream |
                aSelector keywords with:argNames do:[:eachKeyword :eachArgName| 
                    stream nextPutAll:eachKeyword; nextPutAll:eachArgName; space.
                ].
                stream backStep.   "remove the last space"
             ].
        ].
        ^ aSelector , ' ' , (argNames at:1)
    ].
    ^ aSelector

    "
     Method methodDefinitionTemplateForSelector:#foo          andArgumentNames:#()        
     Method methodDefinitionTemplateForSelector:#+            andArgumentNames:#('aNumber') 
     Method methodDefinitionTemplateForSelector:#foo:bar:baz: andArgumentNames:#('fooArg' 'barArg' 'bazArg') 
    "
!

methodPrivacySupported
    "return true, if the system was compiled to support methodPrivacy.
     You should not depend on that feature being available."

%{  /* NOCONTEXT */
#if defined(F_PRIVATE) || defined(F_CLASSPRIVATE)
    RETURN (true);
#else
    RETURN (false);
#endif
%}
!

resourceTypes
    "get the types of resources for which a spec definition and corresponding editor exist"

    ^ #(canvas menu image fileImage help hierarchicalList tabList tableColumns)
! !

!Method class methodsFor:'special'!

flushSourceStreamCache
    LastFileLock critical:[
        LastSourceFileName := LastMethodSources := nil.
        LastFileReference at:1 put:0.
    ].

    "
     Method flushSourceStreamCache
    "

    "Created: 9.2.1996 / 19:05:28 / cg"
! !

!Method methodsFor:'Compatibility-VW'!

classIsMeta
    "return true, if this method is a class method"

    ^ self mclass isMeta 
!

sendsSelector:aSelectorSymbol
    "return true, if this method contains a message-send
     with aSelectorSymbol as selector."

    ^ self sends:aSelectorSymbol
! !


!Method methodsFor:'accessing'!

category
    "return the methods category or nil"

    ^ category
!

category:aStringOrSymbol
    "set the methods category"

    |newCategory oldCategory cls|

    aStringOrSymbol notNil ifTrue:[
        newCategory := aStringOrSymbol.
        newCategory ~= (oldCategory := category) ifTrue:[
            self setCategory:newCategory.

            cls := self mclass.
            cls notNil ifTrue:[
                cls addChangeRecordForMethodCategory:self category:newCategory.
                self changed:#category with:oldCategory.            "/ will vanish
                cls changed:#organization with:self selector.       "/ will vanish
                Smalltalk changed:#methodCategory with:(Array with:cls with:self with:oldCategory). 
            ]
        ]
    ]

    "Modified: / 25-09-2007 / 16:15:24 / cg"
!

comment
    "return the methods comment.
     This is done by searching for and returning the first comment
     from the methods source (excluding any double-quotes). 
     Returns nil if there is no comment (or source is not available)."

    |src comment comments parser|

    src := self source.
    src isNil ifTrue:[^ nil].

    parser := Parser for:src in:nil.
    parser ignoreErrors:true; ignoreWarnings:true; saveComments:true.
    parser parseMethodSpec.

    comments := parser comments.
    comments size ~~ 0 ifTrue:[
        comment := comments first asString.
        (comment withoutSpaces endsWith:'}') ifTrue:[
            "if first comment is a pragma, take next comment"
            comment := comments at:2 ifAbsent:nil.
            comment notNil ifTrue:[
                comment := comment string.
            ].
        ].
    ].
    ^ comment.

    "
     (Method compiledMethodAt:#comment) comment  
     (Object class compiledMethodAt:#infoPrinting:) comment  
    "

    "Modified: / 17.2.1998 / 14:50:00 / cg"
    "Modified: / 23.2.1998 / 10:26:08 / stefan"
!

getPackage
    "return the package-ID of the method"

    ^ package
!

getSource
    "low-level access to the source instance-variable.
     For internal (compiler) use only.
     This is NOT always the methods source string"

    ^ source
!

getSourcePosition
    "low-level access to the sourcePosition instance-variable.
     For internal (compiler) use only.
     This is NOT always the methods sourcePosition"

    ^ sourcePosition
!

localSourceFilename:aFileName position:aNumber
    "set the methods sourcefile/position indicating, that 
     this is a local file."

    source := aFileName.
    sourcePosition := aNumber negated

    "Created: 16.1.1997 / 01:25:52 / cg"
!

makeLocalStringSource
    "assure that the methods source code is stored locally as a string
     within the method (as opposed to an external string, which is accessed
     by reading the source code file).
     This is required, when a methods package is changed, to assure that its
     sourceCode is not lost."

    source notNil ifTrue:[
        sourcePosition notNil ifTrue:[
            self source:(self source)
        ]
    ].
!

mclass:aClass
    "set the method's class"
     
    mclass == aClass ifTrue:[ ^ self ].

"/     (mclass notNil and:[aClass notNil]) ifTrue:[
"/         'Method [warning]: mclass already set' errorPrintCR.
"/     ].
    mclass := aClass.

    "Modified: / 28-11-2006 / 12:12:27 / cg"
!

package
    "return the package-ID of the method (nil is translated to noProject here)"

    |cls|

    package notNil ifTrue:[ ^ package ].
    (cls := self mclass) isNil ifTrue:[
        ^ PackageId noProjectID.
    ].
    "/ set it.
    package := cls getPackage.
    package isNil ifTrue:[
        ^ PackageId noProjectID.
    ].
    ^ package

    "Modified: / 28-11-2006 / 12:12:43 / cg"
!

package:aSymbol
    "set the package-symbol"

    |cls oldPackage newPackage|

    aSymbol == PackageId noProjectID ifTrue:[
        newPackage := nil
    ] ifFalse:[
        newPackage := aSymbol
    ].

    package ~~ newPackage ifTrue:[
        oldPackage := package.
        "/ this is required, because otherwise I would no longer be able to
        "/ reconstruct my sourcecode (as the connection to the source-file is lost).
        self makeLocalStringSource.
        package := newPackage.

        cls := self mclass.

        self changed:#package.                                              "/ will vanish
        cls changed:#methodPackage with:self selector.                      "/ will vanish

        Smalltalk changed:#projectOrganization with:(Array with:cls with:self with:oldPackage). 
        cls addChangeRecordForMethodPackage:self package:newPackage.
    ]

    "Modified: / 23-11-2006 / 17:01:02 / cg"
!

setCategory:aStringOrSymbol
    "set the methods category (without change notification)"

    aStringOrSymbol notNil ifTrue:[
        category := aStringOrSymbol asSymbol
    ]

    "Modified: / 13.11.1998 / 23:55:05 / cg"
!

setPackage:aSymbol
    "set the package-symbol (low level - use package:)"

    package := aSymbol
!

source
    "return the sourcestring for the receiver"

    |aStream chunk|

    "
     if sourcePosition is nonNil, its the fileName and
     abs(sourcePosition) is the offset.
     Otherwise, source is the real source
    "
    sourcePosition isNil ifTrue:[^ source].

    source notNil ifTrue:[
        self class lastMethodSourcesLock critical:[
            LastMethodSources notNil ifTrue:[
                chunk := LastMethodSources at:self ifAbsent:nil.
            ].
        ].
        chunk notNil ifTrue:[
            ^ chunk
        ].

        LastFileLock critical:[
            "have to protect aStream from being closed as a side effect 
             of some other process fetching some the source from a different source file"

            aStream := self sourceStream.
            aStream notNil ifTrue:[
                [
                    chunk := self sourceChunkFromStream:aStream.
                ] on:DecodingError do:[:ex|
                    "CharacterEncoder>>#guessEncoding is not fail safe - retry with plain unencoded data"

                    ('DecodingError ignored when reading <1p> (<2p>)' expandMacrosWith:self whoString with:ex description) infoPrintCR.
                    aStream := self rawSourceStream.
                    ex restart.
                ].
            ].
        ].
    ].

    "Cache the source of recently used methods"
    
    chunk notNil ifTrue:[
        CacheDictionary notNil ifTrue:[
            self class lastMethodSourcesLock critical:[
                LastMethodSources isNil ifTrue:[
                    LastMethodSources := CacheDictionary new:50.
                ].
                LastMethodSources at:self put:chunk.
            ]
        ]
    ].

    ^ chunk

    "Modified: / 07-01-1997 / 16:20:09 / stefan"
    "Modified: / 28-11-2006 / 11:49:25 / cg"
!

source:aString
    "set the methods sourcestring"

    source := aString.
    sourcePosition := nil
!

sourceFilename
    "return the sourcefilename if source is extern; nil otherwise"

    sourcePosition notNil ifTrue:[^ source].
    ^ nil
!

sourceFilename:aFileName position:aNumber
    "set the methods sourcefile/position"

    source := aFileName.
    sourcePosition := aNumber
!

sourceLineNumber
    "return the lineNumber of my source within the returned
     source sourcestring.
     For ST methods, the returned sourceString is always the
     methods pure source; therefore, the lineNumber is always 1."

    ^ 1

    "Created: 30.7.1997 / 15:42:01 / cg"
!

sourcePosition
    "return the sourceposition if source is extern; nil otherwise"

    sourcePosition isNil ifTrue:[^ sourcePosition].
    ^ sourcePosition abs

    "Modified: 16.1.1997 / 01:28:25 / cg"
! !

!Method methodsFor:'accessing-visibility'!

isIgnored
    "return true, if this is an ignored method.
     Ignored methods are physically present in the source file,
     but no code is generated for it by stc, and the VM does not see
     it in its message lookup.
     (i.e. setting a method to #ignored, and sending that selector,
      leads to either the superclasses implementation to be called, 
      or a doesNotUnderstand exception to be raised)

     Notice: this is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ self privacy == #ignored

    "Modified: / 23.1.1998 / 15:23:02 / stefan"
!

isPrivate
    "return true, if this is a private method.
     Execution of private methods is only allowed via self sends
     from superclasses or the class itself.
     If a private method is called by some other class, a runtime
     error (PrivateMethodSignal) is raised.
     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ self privacy == #private

    "Modified: / 23.1.1998 / 15:23:13 / stefan"
!

isProtected
    "return true, if this is a protected method.
     Execution of protected methods is only allowed via self/super sends
     from superclasses, the class itself or subclasse.
     If a protected method is called by some other class, a runtime
     error (PrivateMethodSignal) is raised.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ self privacy == #protected

    "Modified: / 23.1.1998 / 15:23:27 / stefan"
!

isPublic
    "return true, if this is a public method - I.e. can be executed via any send.
     This is the default and how other smalltalk implementations treat all methods.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ self privacy == #public

    "Modified: / 23.1.1998 / 15:23:40 / stefan"
!

isRestricted
    "return the flag bit stating that this method is restricted.
     Execution of the receiver will only be allowed if the system is not in
     'trap restricted mode' (-->ObjectMemory) otherise a runtime 
     error (PrivateMethodSignal) is raised.

     Notice: method restriction is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

%{  /* NOCONTEXT */
#ifdef F_RESTRICTED
    INT f = __intVal(__INST(flags));

    if (f & F_RESTRICTED) {
        RETURN (true);
    }
#endif
%}.
    ^ false
! !

!Method privateMethodsFor:'accessing-visibility'!

primSetPrivacy:aSymbol
    "set the methods access rights (privacy) from a symbol;
     Currently, this must be one of #private, #protected, #public or #ignored.
     #setPrivacy: simply sets the attribute. When changing methods, that 
     have already been called, #privacy: should be used.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     If at all, use it for debugging purposes, to catch messagesends
     which are not supposed to be sent by others.
     (especially, if working in a team, while integrating other peoples work)

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

%{  /* NOCONTEXT */

#if defined(M_PRIVACY)
    INT f = __intVal(__INST(flags));
    INT p;

    if (aSymbol == @symbol(public))
        p = 0;
    else if (aSymbol == @symbol(protected))
        p = F_PRIVATE;
    else if (aSymbol == @symbol(private))
        p = F_CLASSPRIVATE;
    else if (aSymbol == @symbol(ignored))
        p = F_IGNORED;
    else
        RETURN(false);  /* illegal symbol */
        

    f = (f & ~M_PRIVACY) | p;
    __INST(flags) = __mkSmallInteger(f);
#endif

%}.
    ^ true

    "Modified: 27.8.1995 / 22:58:08 / claus"
! !

!Method methodsFor:'accessing-visibility'!

privacy
    "return a symbol describing the methods access rights (privacy);
     Currently, this is one of #private, #protected, #public or #ignored.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     If at all, use it for debugging purposes, to catch messagesends
     which are not supposed to be sent by others.
     (especially, if working in a team, while integrating other peoples work)

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

%{  /* NOCONTEXT */
    /* I made this a primitive to get the define constant from stc.h */

#if defined(M_PRIVACY) && (defined(F_PRIVATE) || defined(F_CLASSPRIVATE) || defined(F_IGNORED))

    INT f = __intVal(__INST(flags));
    switch (f & M_PRIVACY) {

# ifdef F_PRIVATE
    case F_PRIVATE:
        RETURN (@symbol(protected));
        break;
# endif
# ifdef F_CLASSPRIVATE
    case F_CLASSPRIVATE:
        RETURN (@symbol(private));
        break;
# endif
# ifdef F_IGNORED
    case F_IGNORED:
        RETURN (@symbol(ignored));
        break;
# endif
    }
#endif
%}.

    ^ #public
!

privacy:aSymbol
    "set the methods access rights (privacy) from a symbol;
     Currently, this must be one of #private, #protected, #public or #ignored.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     If at all, use it for debugging purposes, to catch message sends
     which are not supposed to be sent by others.
     (especially, if working in a team, while integrating other peoples work)

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    |oldPrivacy|

    oldPrivacy := self privacy.

    (self setPrivacy:aSymbol flushCaches:true) ifTrue:[
        |myClass mySelector|

        myClass := self mclass.
        mySelector := self selector.

        self changed:#privacy.                                       "/ will vanish
        myClass notNil ifTrue:[
            mySelector notNil ifTrue:[
                myClass changed:#methodPrivacy with:mySelector.      "/ will vanish
                Smalltalk changed:#privacyOfMethod with:(Array with:myClass with:self with:oldPrivacy). 
                myClass addChangeRecordForMethodPrivacy:self.
            ]
        ]
    ]

    "Modified: / 23-11-2006 / 17:03:20 / cg"
!

restricted:aBoolean
    "set or clear the flag bit stating that this method is restricted. 
     Execution of the receiver will only be allowed if the system is not in
     'trap restricted mode' (-->ObjectMemory) otherise a runtime
     error (PrivateMethodSignal) is raised.

     Notice: method restriction is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

%{  /* NOCONTEXT */
    /* I made this a primitive to get the define constant from stc.h */

#ifdef F_RESTRICTED
    INT f = __intVal(__INST(flags));
    INT old;

    old = f;
    if (aBoolean == true) 
        f |= F_RESTRICTED;
    else
        f &= ~F_RESTRICTED;
    __INST(flags) = __mkSmallInteger(f);
    if (old & F_RESTRICTED)
        RETURN(true);
#endif
%}.
    ^ false

    "
     (ObjectMemory class compiledMethodAt:#compressingGarbageCollect) restricted:true
    "

    "Created: 7.11.1995 / 20:36:19 / stefan"
!

setPrivacy:aSymbol
    "set the methods access rights (privacy) from a symbol;
     Currently, this must be one of #private, #protected, #public or #ignored.
     #setPrivacy: simply sets the attribute. When changing methods, that 
     have already been called, #privacy: should be used.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     If at all, use it for debugging purposes, to catch messagesends
     which are not supposed to be sent by others.
     (especially, if working in a team, while integrating other peoples work)

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ self setPrivacy:aSymbol flushCaches:true
!

setPrivacy:aSymbol flushCaches:doFlush
    "set the methods access rights (privacy) from a symbol;
     Currently, this must be one of #private, #protected, #public or #ignored.
     #setPrivacy: simply sets the attribute. When changing methods, that 
     have already been called, #privacy: should be used.

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     If at all, use it for debugging purposes, to catch messagesends
     which are not supposed to be sent by others.
     (especially, if working in a team, while integrating other peoples work)

     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    |old sel|

    old := self privacy.
    old == aSymbol ifTrue:[^ false].
    (self primSetPrivacy:aSymbol) ifFalse:[^ false].

    "/
    "/ no need to flush, if changing from private to public
    "/
    doFlush ifTrue:[
        (aSymbol == #public and:[old ~~ #ignored]) ifFalse:[
            (sel := self selector) notNil ifTrue:[
                ObjectMemory flushCachesForSelector:sel
            ] ifFalse:[
                ObjectMemory flushCaches.
            ].
        ].
    ].
    ^ true
! !

!Method methodsFor:'binary storage'!

asByteCodeMethod
    "if the receiver has no bytecodes, create & return a method having
     the same semantics as the receiver, but uses interpreted bytecodes.
     Otherwise, return the receiver. The new method is not installed in
     the methodDictionary of any class - just returned.
     If the method contains primitive code, this may return a method
     without bytecode.
     Can be used to obtain a bytecode version of a machine-code method, 
     for binary storage or dynamic recompilation (which is not yet finished)
     or to compile lazy methods down to executable ones."

    |mthd|

    byteCode notNil ifTrue:[
        "
         is already a bytecoded method
        "
        ^ self
    ].

    ParserFlags
        withSTCCompilation:#never
        do:[
            mthd := self asExecutableMethod.
        ].
    ^ mthd

    "Created: 24.10.1995 / 14:02:32 / cg"
    "Modified: 5.1.1997 / 01:01:53 / cg"
!

asByteCodeMethodWithSource:newSource
    |mthd|

    ParserFlags 
        withSTCCompilation:#never
        do:[
            mthd := self asExecutableMethodWithSource:newSource.
        ].
    ^ mthd

    "Created: 24.10.1995 / 14:02:32 / cg"
    "Modified: 5.1.1997 / 01:01:53 / cg"
!

asExecutableMethod
    "if the receiver has neither bytecodes nor machinecode, create & return a 
     method having semantics as the receivers source. This may be machine code,
     if the system supports dynamic loading of object code and the source includes
     primitive code. However, bytecode is preferred, since it compiles faster.
     Otherwise, return the receiver. The new method is not installed in
     the methodDictionary of any class - just returned.
     Can be used to compile lazy methods down to executable ones."

    |temporaryMethod sourceString|

    byteCode notNil ifTrue:[
        "
         is already a bytecoded method
        "
        ^ self
    ].

    sourceString := self source.
    sourceString isNil ifTrue:[
        'Method [warning]: cannot generate bytecode (no source for compilation)' errorPrintCR.
        ^ nil
    ].

    temporaryMethod := self asExecutableMethodWithSource:sourceString.

    (temporaryMethod isNil or:[temporaryMethod == #Error]) ifTrue:[
        'Method [warning]: cannot generate bytecode (contains primitive code or error)' errorPrintCR.
        ^ nil.
    ].
    "/
    "/ try to save a bit of memory, by sharing the source (whatever it is)
    "/
    temporaryMethod sourceFilename:source position:sourcePosition. 
    ^ temporaryMethod
!

asExecutableMethodWithSource:newSource
    |temporaryMethod cls|

    cls := self containingClass.
    cls isNil ifTrue:[
        'Method [warning]: cannot generate bytecode (no class for compilation)' errorPrintCR.
        ^ nil
    ].

    "we have to sequentialize this using a lock-semaphore,
     to make sure only one method is compiled at a time.
     Otherwise, we might get into trouble, if (due to a timeout)
     another recompile is forced while compiling this one ...
     (happened when autoloading animation demos)
    "
    CompilationLock critical:[
        "
         dont want this to go into the changes file,
         dont want output on Transcript and definitely 
         dont want a lazy method ...
        "
        Class withoutUpdatingChangesDo:[
            |silent lazy|

            silent := Smalltalk silentLoading:true.
            lazy := Compiler compileLazy:false.

            [
                |compiler|

                Class nameSpaceQuerySignal answer:(cls nameSpace)
                do:[
                    compiler := cls compilerClass.

                    "/
                    "/ kludge - have to make ST/X's compiler protocol
                    "/ be compatible to ST-80's
                    "/
                    (compiler respondsTo:#compile:forClass:inCategory:notifying:install:)
                    ifTrue:[
                        temporaryMethod := compiler
                                             compile:newSource
                                             forClass:cls
                                             inCategory:(self category)
                                             notifying:nil
                                             install:false.
                    ] ifFalse:[
                        temporaryMethod := compiler new
                                             compile:newSource 
                                             in:cls 
                                             notifying:nil 
                                             ifFail:nil
                    ].
                ].
            ] ensure:[
                Compiler compileLazy:lazy.
                Smalltalk silentLoading:silent.
            ]
        ].
    ].
    (temporaryMethod isNil or:[temporaryMethod == #Error]) ifTrue:[
        'Method [warning]: cannot generate bytecode (contains primitive code or error)' errorPrintCR.
        ^ nil.
    ].
    "/
    "/ try to save a bit of memory, by sharing the source (whatever it is)
    "/
    temporaryMethod source:newSource. 
    "/
    "/ dont forget the methods class & package ...
    "/
    temporaryMethod setPackage:package.
    temporaryMethod mclass:mclass.
    ^ temporaryMethod
! !

!Method methodsFor:'copying'!

copy
    "redefined to change a source ref into a real string"

    |aCopy|

    aCopy := super copy.
    sourcePosition notNil ifTrue:[
        aCopy source:(self source)
    ].
    aCopy mclass:nil.
    ^ aCopy

    "Modified: 16.1.1997 / 01:27:25 / cg"
! !

!Method methodsFor:'error handling'!

invalidCodeObject
    "{ Pragma: +optSpace }"

    "this method is triggered by the interpreter when a nil or non method
     is about to be executed.
     In this case, the VM sends this to the bad method (the receiver).

     Also, the Compiler creates methods with their code/bytecode set to
     this method if - after a class change - a method cannot be compiled
     and is therefore no longer executable (for example, after an instvar
     has been removed, and a method still tries to access this instvar)

     Thus, we arrive here, when playing around in a classes methodArray,
     or compiler/runtime system is broken :-(, 
     or you ignore the error messages during some recompile."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Modified: 4.11.1996 / 22:45:06 / cg"
!

invalidCodeObjectWith:arg
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 1 arg) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:16:16 / cg"
    "Modified: 4.11.1996 / 22:45:12 / cg"
!

invalidCodeObjectWith:arg with:arg2
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 2 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:16:41 / cg"
    "Modified: 4.11.1996 / 22:45:15 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 3 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:16:51 / cg"
    "Modified: 4.11.1996 / 22:45:18 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 4 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:00 / cg"
    "Modified: 4.11.1996 / 22:45:22 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 5 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:09 / cg"
    "Modified: 4.11.1996 / 22:45:25 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 6 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:17 / cg"
    "Modified: 4.11.1996 / 22:45:28 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 7 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:25 / cg"
    "Modified: 4.11.1996 / 22:45:31 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 8 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:32 / cg"
    "Modified: 4.11.1996 / 22:45:38 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 9 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:37 / cg"
    "Modified: 4.11.1996 / 22:45:41 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 10 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:45 / cg"
    "Modified: 4.11.1996 / 22:45:44 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10 with:arg11
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 11 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:17:52 / cg"
    "Modified: 4.11.1996 / 22:45:47 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10 with:arg11 with:arg12
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 12 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 20:51:28 / cg"
    "Modified: 4.11.1996 / 22:46:01 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10 with:arg11 with:arg12 with:arg13
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 13 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:18:09 / cg"
    "Modified: 4.11.1996 / 22:45:57 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10 with:arg11 with:arg12 with:arg13 with:arg14
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 14 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:18:17 / cg"
    "Modified: 4.11.1996 / 22:45:55 / cg"
!

invalidCodeObjectWith:arg with:arg2 with:arg3 with:arg4 with:arg5 with:arg6 with:arg7 with:arg8 with:arg9 with:arg10 with:arg11 with:arg12 with:arg13 with:arg14 with:arg15
    "{ Pragma: +optSpace }"

    "When recompiling classes after a definition-change, all
     uncompilable methods (with 15 args) will be bound to this method here,
     so that evaluating such an uncompilable method will trigger an error."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseErrorString:'invalid method - not executable'.

    "Created: 4.11.1996 / 21:18:22 / cg"
    "Modified: 4.11.1996 / 22:45:52 / cg"
!

privateMethodCalled
    "{ Pragma: +optSpace }"

    "this error is triggered, if a private or protected method is called.

     If you continue in the debugger, the method will be called,
     and further privacy exceptions will NOT be reported at this call location, 
     until any new method is compiled, or the privacy of any method changes,
     or the caches are flushed.
     (the reason is that after the continue, the method is enterred into the
      calling cache, for which method privacy is not checked.
      Any of the above actions flushes this cache and a privacy check
      is performed again.)
     Future versions may not enter private methods into the cache, to fix this
     (unobvious) behavior. However, then you will get an exception for EVERY
     call to a private method ...

     Notice: method privacy is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    ^ PrivateMethodSignal raiseRequest

    "Modified: 4.11.1996 / 22:46:09 / cg"
!

uncompiledCodeObject
    "{ Pragma: +optSpace }"

    "this method is invoked by methods which contain primitive code,
     but have not been compiled to machine code (either due to an error
     when compiling, or simply because no stc is available.
     For those methods, the compiler generated a method object consisting
     of the original source code, but with this methods machine/byte code."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseRequestWith:self
        errorString:'invalid method - not compiled'.

    "Modified: 4.11.1996 / 22:58:02 / cg"
!

unloadedCodeObject
    "{ Pragma: +optSpace }"

    "this method is invoked by methods which have been binary-unloaded
     For those the source info consists
     of the original source code, but with this methods machine/byte code."

%{
    /*
     * for reasons too far from being explained here,
     * this MUST be a compiled method
     */
%}.
    ^ InvalidCodeError
        raiseRequestWith:self
        errorString:'invalid method - unloaded'.

    "Created: 4.11.1996 / 22:57:54 / cg"
    "Modified: 4.11.1996 / 22:58:28 / cg"
! !


!Method methodsFor:'printing & storing'!

printOn:aStream
    "put a printed representation of the receiver onto aStream.
     Since methods do not store their class/selector, we have to search
     for it here."

    |classAndSelector m wrapped|

    wrapped := false.

    self basicPrintOn:aStream."/ aStream nextPutAll:(self classNameWithArticle).
    aStream nextPut:$(.

    classAndSelector := self who.
    classAndSelector isNil ifTrue:[
        "
         not anchored in any class.
         check if wrapped (to be more informative in inspectors)
        "
        m := self wrapper.
        m notNil ifTrue:[
            classAndSelector := m who.
            wrapped := true.
        ]
    ].
    classAndSelector notNil ifTrue:[
        (classAndSelector methodClass) name printOn:aStream.
        aStream nextPutAll:' '.
        (classAndSelector methodSelector) printOn:aStream.
        wrapped ifTrue:[
            aStream nextPutAll:'; wrapped'
        ].
    ] ifFalse:[
        "
         sorry, a method which is nowhere anchored
        "
        aStream nextPutAll:'unbound'
    ].
    aStream nextPut:$)

    "
     (Object compiledMethodAt:#at:) printOn:Transcript. Transcript cr.
     (Object compiledMethodAt:#at:) copy printOn:Transcript. Transcript cr.

     (Object compiledMethodAt:#at:) whoString printOn:Transcript. Transcript cr.
     (Object compiledMethodAt:#at:) copy whoString printOn:Transcript. Transcript cr.
    "

    "Modified: / 12-10-2006 / 20:32:22 / cg"
!

whoString
    "return a string as className>>selector, if this is not an unbound
     method. Otherwise return 'unbound'. Used with debugging."

    |who|

    who := self who.
    who notNil ifTrue:[
        ^ who methodClass name , ' >> ' , (who methodSelector storeString)
    ].
    ^ 'unboundMethod'

    "
     Method new whoString   
     (Method compiledMethodAt:#whoString) whoString
    "

    "Modified: 1.11.1996 / 16:27:04 / cg"
! !

!Method methodsFor:'private'!

localSourceStream
    "try to open a stream from a local source file,
     searching in standard places."

    |dir fileName aStream|

    package notNil ifTrue:[
        "/
        "/ old: look in 'source/<filename>'
        "/ this is still kept in order to find user-private
        "/ classes in her currentDirectory.
        "/
        fileName := Smalltalk getSourceFileName:(package copyReplaceAll:$: with:$/) , '/' , source.
        fileName notNil ifTrue:[
            aStream := fileName asFilename readStreamOrNil.
            aStream notNil ifTrue:[^ aStream].
        ].
        "/
        "/ new: look in package-dir
        "/
        dir := Smalltalk getPackageDirectoryForPackage:package.
        dir notNil ifTrue:[
            fileName := dir construct:source.
            aStream := fileName asFilename readStreamOrNil.
            aStream notNil ifTrue:[^ aStream].
        ].
    ].
    fileName := Smalltalk getSourceFileName:source.
    fileName notNil ifTrue:[
        aStream := fileName asFilename readStreamOrNil.
    ].
    ^ aStream
!

rawSourceStream
    "return an open sourceStream (needs positioning)"

    |aStream fileName who myClass mgr className classNameSymbol dir mod|

    "
     if sourcePosition is nonNil, its the fileName and
     abs(sourcePosition) is the offset.
     Otherwise, source is the real source
    "
    source isNil ifTrue:[^ nil].
    sourcePosition isNil ifTrue:[^ source readStream].

    (package notNil and:[package ~= PackageId noProjectID]) ifTrue:[
        "/ keep the last source file open, because open/close
        "/ operations maybe slow on NFS-mounted file systems.
        "/ Since the reference to the file is weak, it will be closed
        "/ automatically if the file is not referenced for a while. 
        "/ Neat trick.

        LastFileLock critical:[
            aStream := LastFileReference at:1.
            LastFileReference at:1 put:0.

            (aStream isNil or:[aStream == 0 or:[aStream isOpen not]]) ifTrue:[
                aStream := nil.
            ] ifFalse:[
                LastSourceFileName = (package,'/',source) ifFalse:[
                    aStream close.
                    aStream := nil.
                ]
            ].
            LastSourceFileName := nil.
        ].

        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    "/ a negative sourcePosition indicates
    "/ that this is a local file 
    "/ (not to be requested via the sourceCodeManager)
    "/ This kludge was added, to allow sourceCode to be
    "/ saved to a local source file (i.e. 'st.src')
    "/ and having a clue for which file is meant later.

    sourcePosition < 0 ifTrue:[
        aStream := source asFilename readStreamOrNil.
        aStream isNil ifTrue:[
            "/ search in some standard places
            fileName := Smalltalk getSourceFileName:source.
            fileName notNil ifTrue:[
                aStream := fileName asFilename readStreamOrNil.
            ].
        ].
        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    "/
    "/ if there is no SourceManager, look in local standard places first
    "/
    (Class tryLocalSourceFirst or:[(mgr := Smalltalk at:#SourceCodeManager) isNil]) ifTrue:[
        aStream := self localSourceStream.
        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    "/
    "/ nope - ask my class for the source (this also invokes the SCMgr)
    "/
    who := self who.
    who notNil ifTrue:[
        myClass := who methodClass.

        (package notNil and:[package ~= myClass package]) ifTrue:[
            "/ I am an extension
            mgr notNil ifTrue:[
                "/ try to get the source using my package information ...
                mod := package asPackageId module.
                dir := package asPackageId directory.
                aStream := mgr streamForExtensionFile:source package:package directory:dir module:mod cache:true.
                aStream notNil ifTrue:[
                    ^ aStream
                ].
            ].
            "/ consult the local fileSystem
            aStream := self localSourceStream.
            aStream notNil ifTrue:[
                ^ aStream
            ]
        ].

        aStream := myClass sourceStreamFor:source.
        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    "/
    "/ nope - look in standard places 
    "/ (if there is a source-code manager - otherwise, we already did that)
    "/
    (mgr notNil and:[Class tryLocalSourceFirst not]) ifTrue:[
        aStream := self localSourceStream.
        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    "/
    "/ final chance: try current directory
    "/
    aStream isNil ifTrue:[
        aStream := source asFilename readStreamOrNil.
        aStream notNil ifTrue:[
            ^ aStream
        ].
    ].

    (who isNil and:[source notNil]) ifTrue:[
        "/
        "/ mhmh - seems to be a method which used to be in some
        "/ class, but has been overwritten by another or removed.
        "/ (i.e. it has no containing class anyMore)
        "/ try to guess the class from the sourceFileName.
        "/ and retry.
        "/
        className := Smalltalk classNameForFile:source.
        (classNameSymbol := className asSymbolIfInterned) notNil ifTrue:[
            myClass := Smalltalk at:classNameSymbol ifAbsent:nil.
            myClass notNil ifTrue:[
                aStream := myClass sourceStreamFor:source.
                aStream notNil ifTrue:[
                    ^ aStream
                ].
            ]
        ]
    ].                

    ^ nil

    "Modified: / 26-11-2006 / 22:33:38 / cg"
!

sourceChunkFromStream:aStream
    PositionError handle:[:ex |
        ^ nil
    ] do:[
        aStream position1Based:sourcePosition abs.
    ].
    ^ aStream nextChunk.
!

sourceStream
    "return an open sourceStream (needs positioning)"

    |stream rawStream|

    rawStream := stream := self rawSourceStream.
    stream notNil ifTrue:[
        "/ see if its utf8 encoded...
        stream := ChangeSet decodedStreamFor:stream.
        (package notNil and:[package ~= PackageId noProjectID]) ifTrue:[
            LastFileLock critical:[
                LastSourceFileName := package,'/',source.
                LastFileReference at:1 put:rawStream.
            ]
        ].
    ].
    ^ stream
! !

!Method methodsFor:'private-compiler interface'!

primitiveNumber
    "for stx rel >= 5.x only:
     return the primitive number."

%{  /* NOCONTEXT */

#ifdef F_PRIMITIVE
    INT f = __intVal(__INST(flags));
    OBJ nr = 0;

    if (f & F_PRIMITIVE) {
        nr = __INST(code_);
    }
    RETURN (nr);
#endif
%}.
    self primitiveFailed
!

setPrimitiveNumber:aNumber
    "for stx rel >= 5.x only:
     mark the method as having primitive code."

%{  /* NOCONTEXT */

#ifdef F_PRIMITIVE
    INT f = __intVal(__INST(flags));

    f |= F_PRIMITIVE;
    __INST(flags) = __mkSmallInteger(f);
    __INST(code_) = aNumber;
    RETURN (self);
#endif
%}.
    self primitiveFailed



!

setResourceFlag
    "mark the method as having a <resource> definition in its
     source. This flag can be used to find resource-flagged methods quicker."

%{  /* NOCONTEXT */

#ifdef F_RESOURCE
    INT f = __intVal(__INST(flags));

    f |= F_RESOURCE;
    __INST(flags) = __mkSmallInteger(f);
#endif
%}
! !

!Method methodsFor:'queries'!

accessedInstVars
    "return a collection of instVarNames, which are accessed by
     the receiver method"

    |src parser|

    src := self source.
    src notNil ifTrue:[
        parser := Parser
                        parseMethod:src 
                        in:self containingClass 
                        ignoreErrors:true 
                        ignoreWarnings:true.

        (parser notNil and:[parser ~~ #Error]) ifTrue:[
            ^ parser usedInstVars
        ].
    ].
    ^ #() "/ actually: unknown

    "Modified: 19.6.1997 / 17:54:09 / cg"
!

containingClass
    "return the class I am defined in. 
     See comment in who."

    "based on who, which has been added for ST-80 compatibility"

    |who|

    mclass notNil ifTrue:[
        "/ check if this (cached) info is still valid ...
        (mclass containsMethod:self) ifTrue:[
            ^ mclass
        ].
        mclass := nil.
    ].

    who := self who.
    who notNil ifTrue:[^ who methodClass].
    "
     none found - sorry
    "
    ^ nil

    "
     (Object compiledMethodAt:#at:) containingClass   

     (Object class compiledMethodAt:#version) containingClass   
    "
!

externalLibraryFunction
    "if this is an externalLibraryFunction call, return the externalLibraryFunction.
     Returns nil otherwise."

    ^ self literalsDetect:[:lit | lit isExternalLibraryFunction] ifNone:nil
!

hasPrimitiveCode
    "return true, if the method contains primitive code; false if not.
     Uses Parser to parse methods source and get the information."

    |src|

    "/ first a trivial reject, if the source does not
    "/ contain a '% {' sequence

    src := self source.
    src notNil ifTrue:[
        (src includesString:(String with:$% with:${) "<- no constant here - to avoid trouble with stupid scanners" ) ifFalse:[
            "/ cannot contain primitive code.
            ^ false
        ]
    ].

    "/ ok; it may or may not ...

    ^ self parse:#'parseMethodSilent:' return:#hasPrimitiveCode or:false

    "
     (Method compiledMethodAt:#hasPrimitiveCode) hasPrimitiveCode 
     (Object compiledMethodAt:#at:) hasPrimitiveCode   
     (Object compiledMethodAt:#basicAt:) hasPrimitiveCode 
    "

    "Modified: 22.1.1997 / 00:03:45 / cg"
!

hasResource
    "return true if the method had a <resource> definition in its
     source. This flag can be used to find resource-flagged methods quicker."

%{  /* NOCONTEXT */

#ifdef F_RESOURCE
    RETURN (((INT)__INST(flags) & __MASKSMALLINT(F_RESOURCE)) ? true : false);
#endif
%}.
    ^ false
!

hasResource:aSymbol
    "return true if the method had a <resource> definition for aSymbol."

    ^ self hasResource and:[ self resources includesKey:aSymbol ]

    "
     Method allInstancesDo:[:m |
        (m hasResource:#image) ifTrue:[self halt]
     ].
    "
!

homeMethod
    "for common protocol with blocks: if the receiver is a method,
     return the receiver; otherwise, if its a block, return its home
     method."

    ^ self

    "Created: 19.6.1997 / 16:13:12 / cg"
!

indexOfOLECall
    "return true, if the method contains ole call; false if not.
     Uses Parser to parse methods source and get the information."

    |funcOrNil|

    funcOrNil := self externalLibraryFunction.
    (funcOrNil isNil or:[funcOrNil isCallTypeOLE not]) ifTrue:[^ nil].
    ^ funcOrNil vtableIndex

    "
     (Method compiledMethodAt:#hasPrimitiveCode) isOLECall  
     (Method compiledMethodAt:#hasPrimitiveCode) indexOfOLECall 

     (Win32OperatingSystem class compiledMethodAt:#primClosePrinter:) isOLECall    
     (Win32OperatingSystem class compiledMethodAt:#primClosePrinter:) indexOfOLECall    
     (Win32OperatingSystem class compiledMethodAt:#primClosePrinter:) isExternalLibraryFunctionCall    
     (Win32OperatingSystem class compiledMethodAt:#primClosePrinter:) externalLibraryFunctionCall    

     (IUnknownPointer compiledMethodAt:#invokeAddRef) isExternalLibraryFunctionCall    
     (IUnknownPointer compiledMethodAt:#invokeAddRef) externalLibraryFunction    
     (IUnknownPointer compiledMethodAt:#invokeAddRef) isOLECall    
     (IUnknownPointer compiledMethodAt:#invokeAddRef) indexOfOLECall    
    "
!

isDocumentationMethod
    "Return true, if this is a documentation only (only a comment) method
     (implies being a metaclass method)"

    self mclass isMeta ifFalse:[^ false].
    self isVersionMethod ifTrue:[^ false].
    ^ self parse:#'parseMethodSilent:' return:#isEmptyMethod or:false.
!

isExternalLibraryFunctionCall
    "Return true, if this is an externalLibraryFunction call."

    ^ self externalLibraryFunction notNil
!

isInvalid
    "return true, if this method is not executable due to
     a (re)-compilation error. Since invalidation is by patching the
     methods code-pointer to a trap function, check for that here.
     (see comment in Method>>invalidCodeObject)."

    |m myCode|

    myCode := self code.

    m := self trapMethodForNumArgs:(self numArgs).
    (m notNil and:[self ~~ m]) ifTrue:[
        (myCode notNil and:[myCode = m code]) ifTrue:[^ true].
        (byteCode notNil and:[byteCode == m byteCode]) ifTrue:[^ true].
    ].

    m := Method compiledMethodAt:#uncompiledCodeObject.
    (m notNil and:[self ~~ m]) ifTrue:[
        (myCode notNil and:[myCode = m code]) ifTrue:[^ true].
        (byteCode notNil and:[byteCode == m byteCode]) ifTrue:[^ true].
    ].

    m := Method compiledMethodAt:#unloadedCodeObject.
    (m notNil and:[self ~~ m]) ifTrue:[
        (myCode notNil and:[myCode = m code]) ifTrue:[^ true].
        (byteCode notNil and:[byteCode == m byteCode]) ifTrue:[^ true].
    ].

    ^ false

    "Modified: 4.11.1996 / 23:34:24 / cg"
!

isMethod
    "return true, if the receiver is some kind of method;
     true returned here - the method is redefined from Object."

    ^ true
!

isOLECall
    "return true, if the method is an ole call; false if not."

    |funcOrNil|

    funcOrNil := self externalLibraryFunction.
    ^ funcOrNil notNil and:[funcOrNil isCallTypeOLE]
!

isObsolete
    "returns true, if this method is obsolete and should not be used any longer"

    |res|

    ^ (res := self resources) notNil and:[res includesKey:#obsolete]
!

isVersionMethod
    "Return true, if this is a CVS, SVN or other version method.
     Stupid: need to know all of them here; better add a pragma or
     method attribute for that..."

    ^ self mclass isMeta
    and:[(AbstractSourceCodeManager isVersionMethodSelector:self selector)]

    "
     (Method class compiledMethodAt:#version) isVersionMethod  
     (Method class compiledMethodAt:#documentation) isVersionMethod
    "
!

isVisualWorksTypedef
    "Return true, if this is a type-returning method (a visualWorks typedef)"

    ^ (self literals size > 0)
    and:[ (self literalAt:1) isKindOf:CType ]
!

mclass
    "return the class in which the receiver was compiled.
     Same as #containingClass, for ST80 compatibility."

    ^ self containingClass

    "Created: 19.6.1997 / 16:26:13 / cg"
!

messages
    "return a collection of message-selectors, sent by this method."

    ^ self messagesSent
!

messagesDo:aBlock
    "evaluate aBlock for each message-selector sent by this method.
     Uses Parser to parse methods source and extract the names."

    ^ self messagesSent do:aBlock
!

messagesSent
    "return a collection with the message selectors sent to by the receiver.
     Uses Parser to parse methods source and extract the names.
     The returned collection includes all used message selectors (i.e. including super-send messages)"

    ^ self parse:#'parseMethodSilent:in:' with:self mclass return:#messagesSent or:#() 

    "
     (Method compiledMethodAt:#printOn:) messagesSent 
     (Point compiledMethodAt:#x:) messagesSent 
    "
!

messagesSentToSelf
    "return a collection with the message selectors sent to self by the receiver.
     Uses Parser to parse methods source and extract the names."

    ^ self parse:#'parseMethodSilent:in:' with:self mclass return:#messagesSentToSelf or:#() 

!

messagesSentToSuper
    "return a collection with the message selectors sent to super by the receiver.
     Uses Parser to parse methods source and extract the names."

    ^ self parse:#'parseMethodSilent:in:' with:self mclass return:#messagesSentToSuper or:#() 

!

methodArgAndVarNames
    "return a collection with the methods argument and variable names.
     Uses Parser to parse methods source and extract the names.
     Returns nil if the source is not available, or some other
     syntax/parse error occurred. For methods with no args and no vars,
     an empty collection is returned."

    |parserClass parser sourceString argNames varNames|

    parserClass := self parserClass.
    sourceString := self source.
    (parserClass notNil and:[sourceString notNil]) ifTrue:[
        parser := parserClass parseMethodArgAndVarSpecificationSilent:sourceString.
        (parser isNil or:[parser == #Error]) ifTrue:[^ nil].
        argNames := parser methodArgs.
        varNames := parser methodVars.
        argNames isNil ifTrue:[^ varNames].
        varNames isNil ifTrue:[^ argNames].
        ^ (argNames , varNames)
    ].
    ^ nil

    "
     (Method compiledMethodAt:#printOn:) methodArgAndVarNames
    "
!

methodArgNames
    "return a collection with the methods argument names.
     Uses Parser to parse methods source and extract the names."

    ^ self parse:#'parseMethodSpecificationSilent:' return:#methodArgs or:nil

    "
     (Method compiledMethodAt:#printOn:) methodArgNames
    "

    "Modified: 31.10.1995 / 14:36:46 / cg"
!

methodComment
    "return the methods first comment, nil if there is none.
     This is a somewhat stupid implementation."

    |text line nQuote index qIndex qIndex2 comment|

    text := self source asCollectionOfLines.
    (text size < 2) ifTrue:[^nil].

    line := (text at:2).
    nQuote := line occurrencesOf:(Character doubleQuote).
    (nQuote == 2) ifTrue:[
        qIndex := line indexOf:(Character doubleQuote).
        qIndex2 := line indexOf:(Character doubleQuote) startingAt:(qIndex + 1).
        ^ line copyFrom:(qIndex + 1) to:(qIndex2 - 1)
    ].
    (nQuote == 1) ifTrue:[
        qIndex := line indexOf:(Character doubleQuote).
        comment := line copyFrom:(qIndex + 1).
        (line indexOf:$/ startingAt:qIndex) == (qIndex+1) ifTrue:[
            "/ an EOL comment
            ^ (comment copyFrom:2) withoutSeparators
        ].

        "/ not an EOL comment
        index := 3.
        line := text at:index.
        nQuote := line occurrencesOf:(Character doubleQuote).
        [nQuote ~~ 1] whileTrue:[
            comment := comment , Character cr asString , line withoutSpaces.
            index := index + 1.
            line := text at:index.
            nQuote := line occurrencesOf:(Character doubleQuote)
        ].
        qIndex := line indexOf:(Character doubleQuote).
        ^ comment , Character cr asString , (line copyTo:(qIndex - 1)) withoutSpaces
    ].
    ^ nil

    "
     (Method compiledMethodAt:#methodComment) methodComment
    "
!

methodDefinitionTemplate
    "return the string that defines the method and the arguments"

    ^ Method methodDefinitionTemplateForSelector:self selector andArgumentNames:self methodArgNames

    "
      (self compiledMethodAt:#printOn:) methodDefinitionTemplate
    "
!

methodVarNames
    "return a collection with the methods local-variable names.
     Uses Parser to parse methods source and extract the names."

    ^ self parse:#'parseMethodArgAndVarSpecificationSilent:' return:#methodVars or:nil

    "
     (Method compiledMethodAt:#printOn:) methodVarNames
    "

    "Modified: 31.10.1995 / 14:36:49 / cg"
!

modificationTime
    "try to extract the modificationTime as a timeStamp from
     the receivers source. If there is no source or no history line, 
     we do not know the modification time, and nil is returned."

    |s list histLine|

    HistoryManager isNil ifTrue:[^ nil].

    s := self source.
    s isNil ifTrue:[^ nil].
    list := HistoryManager getAllHistoriesFrom:s.
    list size == 0 ifTrue:[^ nil].
    histLine := list last.
    ^ Timestamp 
        fromDate:histLine date 
        andTime:histLine time

    "
     (Method compiledMethodAt:#modificationTime) modificationTime
     (Method compiledMethodAt:#isMethod) modificationTime 
    "

    "Modified: 8.9.1995 / 15:08:22 / claus"
    "Modified: 4.11.1996 / 22:28:17 / cg"
!

name
    "for compatibility with javaMethods"

    ^ self selector

    "Created: / 9.11.1998 / 06:15:08 / cg"
!

parse:parseSelector return:accessSelector or:valueIfNoSource 
    "helper for methodArgNames, methodVarNames etc.
     Get the source, let parser parse it using parseSelector,
     return parser-info using accessSelector"

    ^ self parse:parseSelector with:nil return:accessSelector or:valueIfNoSource

    "
     (Method compiledMethodAt:#parse:return:or:)
        parse:#'parseMethodSilent:' return:#sentMessages or:#() 
    "
!

parse:parseSelector with:arg2 return:accessSelector or:valueIfNoSource 
    "helper for methodArgNames, methodVarNames etc.
     Get the source, let parser parse it using parseSelector,
     return parser-info using accessSelector"

    |parser parserClass sourceString|

    parserClass := self parserClass.
    sourceString := self source.
    (parserClass notNil and:[sourceString notNil]) ifTrue:[
        parseSelector numArgs == 2 ifTrue:[
            parser := parserClass perform:parseSelector with:sourceString with:arg2.
        ] ifFalse:[
            parser := parserClass perform:parseSelector with:sourceString.
        ].
        (parser isNil or:[parser == #Error]) ifTrue:[^ valueIfNoSource].
        ^ parser perform:accessSelector
    ].
    ^ valueIfNoSource

    "
     (Method compiledMethodAt:#parse:return:or:)
        parse:#'parseMethodSilent:' return:#sentMessages or:#() 
    "
!

parseResources
    "return the methods resource spec; either nil or a collection of symbols."

    |src parser|

    src := self source.
    src isNil ifTrue:[
        ^ nil "/ actually: dont know
    ].

    (src findString:'resource:') == 0 ifTrue:[
        ^ nil "/ actually: error
    ].
    "/ no need to parse all - only interested in resource-info
    self parserClass isNil ifTrue:[
        ^ nil 
    ].
    parser := self parserClass parseMethodArgAndVarSpecificationSilent:src in:nil.
    parser isNil ifTrue:[
        ^ nil "/ actually error
    ].
    ^ parser primitiveResources.
!

previousVersion
    "return the receivers previous versions source code"

    |sel cls previous|

    sel := self selector.
    sel isNil ifTrue:[ ^ nil ].

    cls := self mclass.
    cls isNil ifTrue:[ ^ nil ].

    ChangeSet current reverseDo:[:change | 
        (change isMethodChange 
        and:[ (change selector == sel)
        and:[ change changeClass == cls ]]) 
        ifTrue:[
            previous := change previousVersion.
            previous notNil ifTrue:[
                ^ previous
            ]
        ]
    ].
    ^ nil.

"/    history := Class methodHistory.
"/    history isNil ifTrue:[^ nil].
"/
"/    entry := history detect:[:entry | |type old new|
"/                                    type := entry first.
"/                                    type == #methodChange ifTrue:[
"/                                        old := entry second.
"/                                        new := entry third.
"/                                        new == self
"/                                    ] ifFalse:[
"/                                        false
"/                                    ]
"/                             ] 
"/                     ifNone:nil.
"/    entry isNil ifTrue:[^nil].
"/    ^ entry second.
"/    ^ history at:self ifAbsent:nil

    "Modified: 7.11.1996 / 19:06:22 / cg"
!

previousVersionCode
    "return the receivers previous versions source code"

    |previous|

    previous := self previousVersion.
    previous isNil ifTrue:[ ^ nil ].

    ^ previous source
!

previousVersions
    "return a collection of the receivers previous versions (sources)"

    |sel cls versions lastChange firstSrc last|

    sel := self selector.
    sel isNil ifTrue:[^ #()].

    cls := self mclass.
    cls isNil ifTrue:[^ #()].

    versions := OrderedCollection new.

    ChangeSet current reverseDo:[:change | 
         (change isMethodChange 
        and:[ (change selector == sel)
        and:[ change changeClass == cls ]]) 
        ifTrue:[
            versions addFirst:change.
            lastChange := change.    
        ]
    ].

    lastChange notNil ifTrue:[
        last := lastChange previousVersion.
        last notNil ifTrue:[
            firstSrc := last source.
            (firstSrc size > 0
            and:[ firstSrc ~= lastChange source]) ifTrue:[
                versions addFirst:(MethodChange 
                                    className:lastChange className 
                                    selector:lastChange selector
                                    source:firstSrc
                                    category:lastChange category).
            ]
        ]
    ].
    ^ versions
!

readsField:instVarIndex
    "return true, if the instvar at instVarIndex is read by the receiver.
     Uses parser (for now); could look at bytecode as well here..."

    |varName readInstVars|

    varName := (self mclass allInstVarNames) at:instVarIndex.
    readInstVars := self parse:#'parseMethodSilent:in:' with:self mclass  return:#readInstVars or:#().
    ^ readInstVars includes:varName.
!

resourceType
    "ST-80 compatibility:
     return the methods first resource specs key. 
     Returns either nil, or a single symbol."

    |resources|

    (resources := self resources) notNil ifTrue:[
        resources keysAndValuesDo:[:key :val|
            ^ key
        ].
    ].
    ^ nil
!

resources
    "return the methods resource spec; either nil or a collection of symbols."

    self hasResource ifFalse:[^ nil].
    ^ self parseResources.
!

selector
    "return the selector under which I am found in my containingClasses
     method-table. 
     See comment in who."

    "based on who, which has been added for ST-80 compatibility"

    |who|

    who := self who.
    who notNil ifTrue:[^ who methodSelector].
    "
     none found - sorry
    "
    ^ nil

    "
     |m|
     m := Object compiledMethodAt:#at:.
     m selector
    "

    "Modified: 1.11.1996 / 16:25:48 / cg"
!

sends:aSelectorSymbol
    "return true, if this method contains a message-send
     with aSelectorSymbol as selector."

    (self referencesLiteral:aSelectorSymbol) ifTrue:[
        ^ self messagesSent includesIdentical:aSelectorSymbol
    ].
    ^ false
!

shouldBeSkippedInDebuggersWalkBack
    "return true, if this method thinks, it should be skipped in a walkback.
     This is done by looking for a #skipInDebuggersWalkBack flag in the methods resources."

    |resources|

    resources := self resources.
    ^ resources notNil and:[ resources includesKey:#skipInDebuggersWalkBack ].
!

superMessages
    "return a collection of message-selectors, sent to super by this method."

    ^ self messagesSentToSuper
!

usedGlobals
    "return a collection with the global names referred to by the receiver.
     Uses Parser to parse methods source and extract them."

    ^ self parse:#'parseMethodSilent:in:' with:self mclass return:#usedGlobals or:#() 

    "
     (Method compiledMethodAt:#resources) usedGlobals 
    "
!

usedSymbols
    "return a collection with the symbols referred to by the receiver.
     Uses Parser to parse methods source and extract them.
     This collection only includes implicit symbols references 
     (i.e. not messages sent)"

    ^ self parse:#'parseMethodSilent:in:' with:self mclass return:#usedSymbols or:#() 

    "
     (Method compiledMethodAt:#usedSymbols) usedSymbols 
     (Method compiledMethodAt:#usedSymbols) messagesSent 
    "
!

who
    "return the class and selector of where I am defined in;
     nil is returned for unbound methods.

     ST/X special notice: 
        returns an instance of MethodWhoInfo, which
        responds to #methodClass and #methodSelector query messages.
        For backward- (& ST-80) compatibility, the returned object also
        responds to #at:1 and #at:2 messages.

     Implementation notice:
        Since there is no information of the containing class 
        in the method, we have to do a search here.

        Normally, this is not a problem, except when a method is
        accepted in the debugger or redefined from within a method
        (maybe done indirectly, if #doIt is done recursively)
        - the information about which class the original method was 
        defined in is lost in this case.

     Problem: 
        this is heavily called for in the debugger to create
        a readable context walkback. For unbound methods, it is
        slow, since the search (over all classes) will always fail.

     Q: should we add a backref from the method to the class 
        and/or add a subclass of Method for unbound ones ?
     Q2: if so, what about the bad guy then, who copies methods around to
         other classes ?"

    |classes cls sel fn clsName checkBlock|

    mclass notNil ifTrue:[
        sel := mclass selectorAtMethod:self.
        sel notNil ifTrue:[
            ^ MethodWhoInfo class:mclass selector:sel
        ].
        "/ flush outdated mclass info
        mclass := nil.
    ].

    checkBlock := [:cls |
        |sel|

        sel := cls selectorAtMethod:self.
        sel notNil ifTrue:[
            LastWhoClass := cls theNonMetaclass name.
            mclass isNil ifTrue:[
                mclass := cls
            ].
            ^ MethodWhoInfo class:cls selector:sel
        ].
    ].

    "
     speedup kludge: if my sourceFileName is valid,
     extract the className from it and try that class first.
    "
    (fn := self sourceFilename) notNil ifTrue:[
        clsName := fn asFilename nameWithoutSuffix.
        clsName := clsName asSymbolIfInterned.
        clsName notNil ifTrue:[
            cls := Smalltalk at:clsName ifAbsent:nil.
            cls notNil ifTrue:[
                checkBlock value:cls theNonMetaclass.
                checkBlock value:cls theMetaclass.
            ]
        ].
    ].

    "
     then, look in the class we found something the last time
     this may often give a hit, when asking who repeatingly for
     a context chain. (keep last by its name, to not keep classes from
     being garbage collected)
    "
    LastWhoClass notNil ifTrue:[
        cls := Smalltalk at:LastWhoClass ifAbsent:nil.
        cls notNil ifTrue:[
            checkBlock value:cls theNonMetaclass.
            checkBlock value:cls theMetaclass.
        ]
    ].

    "
     first, limit the search to global classes only - 
     since probability is high, that the receiver is found in there ...
    "
    classes := Smalltalk allClasses.
    "
     instance methods are usually more common - search those first
    "
    classes do:[:cls |
        checkBlock value:cls theNonMetaclass.
        checkBlock value:cls theMetaclass.
    ].

    LastWhoClass := nil.
"/    "
"/     mhmh - must be a method of some anonymous class (i.e. one not
"/     in the Smalltalk dictionary). Search all instances of Behavior
"/    "
"/    Behavior allSubInstancesDo:[:someClassLikeThingy |
"/        (classes includes:someClassLikeThingy) ifFalse:[
"/            checkBlock value:someClassLikeThingy.
"/        ]
"/    ].
    "
     none found - sorry
    "
    ^ nil

    "typical situation: some well-known class"
    "
     |m|
     m := Object compiledMethodAt:#copy.
     m who
    "

    "untypical situation: an anonymous class"
    "
     |m cls|

     Object 
        subclass:#FunnyClass 
        instanceVariableNames:'foo'
        classVariableNames:''
        poolDictionaries:''
        category:'testing'.
     cls := Smalltalk at:#FunnyClass.
     Smalltalk removeClass:cls.

     cls compile:'testMethod1:arg foo:=arg'.
     cls compile:'testMethod2 ^ foo'.
     m := cls compiledMethodAt:#testMethod1:.

     m who
    "

    "Modified: / 07-11-2006 / 13:58:50 / cg"
!

wrapper
    "only for wrapped methods: return the wrapper.
     Thats the WrapperMethod which contains myself."

    WrappedMethod allInstancesDo:[:m |
        m originalMethod == self ifTrue:[^ m].
    ].
    ^ nil
!

writesField:instVarIndex
    "return true, if the instvar at instVarIndex is written (modified) by the receiver.
     Uses parser (for now); could look at bytecode as well here..."

    |varName modifiedInstVars|

    varName := (self mclass allInstVarNames) at:instVarIndex.
    modifiedInstVars := self parse:#'parseMethodSilent:in:' with:self mclass return:#modifiedInstVars or:#().
    ^ modifiedInstVars includes:varName.
! !

!Method methodsFor:'trap methods'!

makeInvalid
    "make the receiver an invalid method, which raises an invalidCodeObject
     signal when executed. This is not for public use - it is required for
     the objectFileLoader to invalidate methods whose code is unloaded."

    |invldMethod|

    invldMethod := self trapMethodForNumArgs:(self numArgs).
    self code:invldMethod code.
    self byteCode:nil.

    "Created: 17.9.1995 / 15:00:52 / claus"
    "Modified: 4.11.1996 / 23:04:34 / cg"
!

makeUncompiled
    "make the receiver an uncompiled method, which raises an invalidCodeObject
     signal when executed. This is not for public use - it is required for
     the compiler to invalidate methods which cannot be compiled due to errors
     after a class definition change (for example: instvars are no longer there)."

    |invldMethod|

    invldMethod := self class compiledMethodAt:#uncompiledCodeObject.
    self code:invldMethod code.
    self byteCode:nil.

    "Created: 17.9.1995 / 15:01:14 / claus"
    "Modified: 4.11.1996 / 23:04:46 / cg"
!

makeUnloaded
    "make the receiver an unloaded method, which raises an invalidCodeObject
     signal when executed. This is not for public use - it is required for
     the objectFileLoader to invalidate methods for which a shared library has
     been removed."

    |invldMethod|

    invldMethod := self class compiledMethodAt:#unloadedCodeObject.
    self code:invldMethod code.
    self byteCode:nil.

    "Created: 4.11.1996 / 22:59:21 / cg"
    "Modified: 4.11.1996 / 23:04:52 / cg"
!

trapMethodForNumArgs:numArgs
    |trapSel|

    trapSel := #(
                  #'invalidCodeObject'
                  #'invalidCodeObjectWith:'
                  #'invalidCodeObjectWith:with:'
                  #'invalidCodeObjectWith:with:with:'
                  #'invalidCodeObjectWith:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:with:with:with:with:'
                  #'invalidCodeObjectWith:with:with:with:with:with:with:with:with:with:with:with:with:with:with:'
                ) at:(numArgs + 1).

    ^ Method compiledMethodAt:trapSel.

    "Created: 4.11.1996 / 21:58:58 / cg"
    "Modified: 4.11.1996 / 23:18:05 / cg"
! !

!Method::MethodWhoInfo class methodsFor:'documentation'!

documentation
"
    In earlier times, Method>>who returned an array filled
    with the methods class and selector.
    This was done, since a smalltalk method cannot return multiple
    values, but 2 values had to be returned from that method.
    Thus, the who-interface was used as:
        info := <someMethod> who.
        class := info at:1.
        sel := info at:2.

    Sure, this is ugly coding style, and the system has been changed to return
    an object (an instance of MethodWhoInfo) which responds to the two
    messages: #methodClass and #methodSelector.
    This allows to write things much more intuitive:
        info := <someMethod> who.
        class := info methodClass.
        sel := info methodSelector.

    However, to be backward compatible, the returned object still responds to
    the #at: message, but only allows inices of 1 and 2 to be used.

    The MethodWhoInfo class is private to Method - its not visible to other
    classes.

    [author:]
        Claus Gittinger

    [see also:]
        Method
"
! !

!Method::MethodWhoInfo class methodsFor:'instance creation'!

class:cls selector:sel
    "return a new MethodWhoInfo object;
     this is a private interface for Method"

    ^ self basicNew class:cls selector:sel

    "Modified: 2.1.1997 / 15:27:59 / cg"
! !

!Method::MethodWhoInfo methodsFor:'accessing'!

method
    ^ myClass compiledMethodAt:mySelector
!

methodClass
    "return the class which contains the method represented by myself"

    ^ myClass

    "
     (Method compiledMethodAt:#who) who methodClass
     (Method::MethodWhoInfo compiledMethodAt:#methodClass) who methodClass
    "

    "Modified: 2.1.1997 / 14:59:02 / cg"
!

methodSelector
    "return the selector under which the the method represented by myself
     is found in the class"

    ^ mySelector

    "Modified: 2.1.1997 / 14:59:24 / cg"
! !

!Method::MethodWhoInfo methodsFor:'compatibility'!

at:index
    "simulate the old behavior (when Method>>who returned an array)"

    index == 1 ifTrue:[
        ^ myClass
    ].
    index == 2 ifTrue:[
        ^ mySelector
    ].

    "/ sigh - full compatibility ?
    ^ self indexNotIntegerOrOutOfBounds:index
! !

!Method::MethodWhoInfo methodsFor:'private-accessing'!

class:cls selector:sel
    myClass := cls.
    mySelector := sel


! !

!Method class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Method.st,v 1.334 2009-10-22 12:02:09 cg Exp $'
!

version_CVS
    ^ '$Header: /cvs/stx/stx/libbasic/Method.st,v 1.334 2009-10-22 12:02:09 cg Exp $'
! !

Method initialize!
